/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#line 371 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#define XK_MISCELLANY
#include <X11/keysymdef.h>

#include "x-basics.h"
#include "x-props.h"
#include "x-events.h"
#include "x-misc-props.h"
#include "x-image.h"
#include "txtren.h"
#include "config.h"
#include "clients.h"
#include "graymap.h"
#include "alttab.h"

#line 17 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"

//static unsigned int WS_border    =   0x000000;
static unsigned int WS_bg        =   0xefefef;
static unsigned int WS_fg        =   0x000000;
static unsigned int WS_hilight   =   0x000000;
static          int WS_pad_top   =   10;
static          int WS_pad_mid   =   6;
static          int WS_pad_bot   =   10;
static          int WS_pad_horz  =   10;
static          int WS_pad_icon  =   10;
static          int WS_icon_size =   16;
static          int WS_width     =   320;
static          int WS_font_size =   16;
static const char  *WS_font_name =   ":default";
static       char  *WS_keyspec   =   "alt+Tab";
static          int WS_on        =   0;


static u_font_t *WSfont;
static Window    WSwindows[50];
static int       WSn_windows;
static Window    WSwindow;
static int       WSwindow_w;
static int       WSwindow_h;
static ximage_t *WSimage;
static int       WSitem_h;
static int       WScurrent;
static int       WSkey;
static int       WSmod;
static int       WSprev;


#line 57 "alttab.c"

static int get_switcher_clients(client_window_t **clients,int max);
static void render_switcher_item_icon(int y,client_window_t *cw);
static void render_switcher_item_text(int y,client_window_t *cw);
static void render_switcher_item(int y,client_window_t *cw);
static void switcher_expose(void *ignore,XExposeEvent *ev);
static int client_in_list
   (Window window, client_window_t **clients,int n_clients);
static void client_list_rotate_left
  (client_window_t **cl, int n_clients, int rot);
static void switcher_hilight(int idx,int color,int clr);
static void alttab_press(void *arg,XEvent *event);
static void alttab_release(void *arg,XEvent *event);
static void keypress(void *arg, XEvent *event);
static void keyrelease(void *arg,XEvent *event);
#line 53 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"

static int get_switcher_clients(client_window_t **clients,int max)
{
  Window *windows;
  size_t n_windows;
  int i;
  int n_clients;
  uint32_t c_desk;

  if (get_window_property_windows(x_root_window,
                  a_net_client_list_stacking,&n_windows, &windows,0))
    return 0;

  if (get_current_desktop(&c_desk)) c_desk= 0;

  n_clients= 0;
  for(i=0;n_clients<max && i<n_windows;i++)
  {
    client_window_t *C;
    C= get_client(windows[i]);
    if (!C) continue;
    if (!C || get_client_type(C) || C->type!=0) continue;
    if (!get_client_desktop_number(C) && C->desk!=c_desk) continue;
    clients[n_clients++]= C;
  }
  free(windows);
  return n_clients;
}

static void render_switcher_item_icon(int y,client_window_t *cw)
{
  int icon_sx,icon_sy,icon_cw,icon_ch,icon_dx,icon_dy;
  int delta;

  if (get_client_icon(cw,WS_icon_size,WS_icon_size))
    return ;

  delta= cw->icon_width - WS_icon_size;

  if (delta>=0)
  {
    icon_sx= delta/2;
    icon_cw= WS_icon_size;
    icon_dx= 0;
  }
  else
  {
    icon_sx= 0;
    icon_cw= cw->icon_width;
    icon_dx= -delta/2;
  }

  delta= cw->icon_height - WS_icon_size;
  if (delta>=0)
  {
    icon_sy= delta/2;
    icon_ch= WS_icon_size;
    icon_dy= 0;
  }
  else
  {
    icon_sy= 0;
    icon_ch= cw->icon_height;
    icon_dy= -delta/2;
  }

  blend_alpha_image(WSimage->data,WSwindow_w,WSwindow_h, 
                     WS_pad_horz+icon_dx, y+ (WSitem_h-icon_ch)/2+icon_dy,
                     cw->icon_data, cw->icon_width, cw->icon_height,
                     icon_sx, icon_sy, icon_cw, icon_ch, WS_bg);
}

static void render_switcher_item_text(int y,client_window_t *cw)
{
  char *str;
  uint32_t titlew;
  uint8_t *title;
  int copyw;
  int dy;

  if (get_client_title(cw)) str= "Window"; else str= cw->title;
  text_render(WSfont, str, &titlew, &title);
  copyw= WSwindow_w - (WS_pad_horz*2+WS_icon_size+WS_pad_icon);
  dy= (WSitem_h - WSfont->height)/2;
  graymap_draw(WSimage->data, WSwindow_w, WSwindow_h,
                  WS_pad_horz + WS_icon_size + WS_pad_icon, y + dy, 
               title, titlew, WSfont->height, 0,0, WS_fg, WS_bg,
                  copyw, WSfont->height);
  free(title);
}

static void render_switcher_item(int y,client_window_t *cw)
{
  render_switcher_item_icon(y,cw);
  render_switcher_item_text(y,cw);
}

static void switcher_expose(void *ignore,XExposeEvent *ev)
{
  put_image(WSwindow, ev->x, ev->y, 
            WSimage, ev->x,ev->y, ev->width,ev->height);
}

#line 157 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"
static int client_in_list
   (Window window, client_window_t **clients,int n_clients)
{
  int i;
  for(i=0;i<n_clients;i++)
    if (clients[i]->id==window)
      return i;
  return -1;
}

static void client_list_rotate_left
  (client_window_t **cl, int n_clients, int rot)
{
  int i;
  client_window_t *tmp[50];
  for(i=0;i<n_clients;i++)
    tmp[i]= cl[ (i+rot)%n_clients ];
  for(i=0;i<n_clients;i++)
    cl[i]= tmp[i];
}

#line 250 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"

static void switcher_hilight(int idx,int color,int clr)
{
  int y;
  int sx,ex, sy,ey;
  y= WS_pad_top+idx*(WSitem_h+WS_pad_mid);

  sy= y - WS_pad_mid/2;
  ey= sy + WSitem_h+WS_pad_mid;
  sx= WS_pad_horz / 2;
  ex= WSwindow_w - WS_pad_horz/2;

  draw_rect(WSimage->data,WSwindow_w,WSwindow_h,
            sx, sy, ex-sx, ey-sy, color ? WS_hilight : WS_bg);

  if (clr)
    XClearArea(x_display, WSwindow,sx,sy,ex-sx,ey-sy,True);
}

#line 308 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"

static void alttab_press(void *arg,XEvent *event)
{
#if 0
  printf("Press\n");
#endif
  if (grab_keyboard(keypress, keyrelease, NULL))
   { printf("keyboard grab failed\n"); }
  else { switcher_begin(); }
}

static void alttab_release(void *arg,XEvent *event)
{
#if 0
  printf("Release\n");
#endif
}

static void keypress(void *arg, XEvent *event)
{
  if (event->xkey.keycode==WSkey) 
  {
     switcher_step();
  }
  else
  {
#if 0
    printf("Grabbed key press\n");
#endif
  }
}

static void keyrelease(void *arg,XEvent *event)
{
  KeySym sym;
#if 0
  printf("grabbed key release\n");
#endif
  sym= XLookupKeysym(&event->xkey, 0);
  if (sym==NoSymbol) 
  {
     static int n;
     n++;
     if (n==20) { n=0; goto release; }
     return ;
  }
  if (sym!=XK_Alt_L && sym!=XK_Alt_R)
    return ;
release:
#if 0
  printf("ungrabbing the keyboard\n");
#endif
  ungrab_keyboard();
  switcher_end();
}

#line 179 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"

void switcher_begin()
{
  int i,y;
  Window active_window;
  client_window_t *clients[50];
  int n_clients;

  if (!WS_on) return ;

  active_window= None;
  WSimage= NULL;

  n_clients= get_switcher_clients(clients,50);
  if (n_clients<2) return ;

  if (!WSfont)
    WSfont= font_new(WS_font_name,WS_font_size);

  WSwindow_w= WS_width;
  WSitem_h= WSfont->height;
  if (WSitem_h < WS_icon_size) WSitem_h= WS_icon_size;
  WSwindow_h= n_clients * WSitem_h + (n_clients-1)*WS_pad_mid
               + WS_pad_top+WS_pad_bot;

  if (!get_active_window(&active_window))
  {
    int offset;
    offset= client_in_list(active_window, clients, n_clients);
    if (offset>0) 
       client_list_rotate_left(clients, n_clients, offset);
  }

  
  if (active_window!=WSprev)
  {
    i= client_in_list(WSprev, clients, n_clients);
    if (i>=0 && i!=1)
    {
      client_window_t *tmp;
      tmp= clients[1];
      clients[1]= clients[i];
      clients[i]= tmp;
    }
  }

  for(i=0;i<n_clients;i++)
    WSwindows[i]= clients[i]->id;
  WSn_windows= n_clients;

  WSimage= create_image(WSwindow_w,WSwindow_h);
  WSwindow= create_window_Ob(x_root_window, 
                             (x_root_width-WSwindow_w)/2,
                             (x_root_height-WSwindow_h)/2,
                             WSwindow_w,WSwindow_h, ExposureMask, 3);
  fill_rect(WSimage->data,WSwindow_w,WSwindow_h,0,0,
            WSwindow_w,WSwindow_h,WS_bg);
  y= WS_pad_top;
  for(i=0;i<n_clients;i++)
  {
    render_switcher_item(y, clients[i]);
    y+= WSitem_h + WS_pad_mid;
  }

  register_event_handler(WSwindow, Expose, NULL, switcher_expose, NULL);
  WScurrent= 1;
  switcher_hilight(1,1,0);
  WSprev= active_window;
}

#line 270 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"

void switcher_step()
{
  if (!WSimage) return; 
  switcher_hilight(WScurrent, 0,1);
  WScurrent= (WScurrent+1)%WSn_windows;
#if 0
printf("nwindows: %d current= %d\n", WSn_windows, WScurrent);
#endif
  switcher_hilight(WScurrent, 1,1);
}

void switcher_end()
{
  if (!WSimage) return ;

  unregister_event_handler(WSwindow, Expose, NULL);
  destroy_image(WSimage);
  WSimage= NULL;
  XDestroyWindow(x_display,WSwindow);

  if (WScurrent<0 || WScurrent>=WSn_windows) return ;
  set_active_window(WSwindows[WScurrent]);
}

void switcher_init(cfg_switcher_t *cfg)
{
#define OP(x) if (cfg->G_##x) WS_##x= cfg->x;
  OP(fg) OP(bg) OP(hilight) OP(pad_top) OP(pad_mid) OP(pad_bot) 
  OP(pad_horz) OP(pad_icon) OP(width) OP(font_size) OP(font_name)
#undef OP
  if (cfg->G_enabled) WS_on= cfg->enabled;
  if (!WS_on) return ;
  if (grab_key(WS_keyspec,&WSmod, &WSkey, alttab_press, alttab_release, NULL))
     WS_on= 0; 
}

#line 399 "/home/dodo/tmp/mupanel-orig-1.5/src/alttab.u"

