/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include "appmenu.h"

static void read_filep_real(FILE *f,char**buf, size_t *len)
{
    char *xb;
    static const int rsize=2000;
    size_t r, pos= *len;
    int eno;

    xb= malloc(rsize);
    r= fread(xb, 1, rsize, f);
    *len+= r;
    if (r==rsize)
    {
        read_filep_real(f, buf,len);
    }
    else
    {
       if (ferror(f)) 
       {
         eno= errno;
         free(xb);
         errno= eno;
         *buf= 0;
         return ;
       }
       *buf= malloc(*len+1);
    }
        
    memcpy(*buf + pos, xb, r);
    free(xb);
}

static void read_filep(FILE *f,char**buf, size_t *len) {
    *len= 0;
    read_filep_real(f,buf,len);
    if (*buf)
       (*buf)[*len]= 0;
}

static char* file_read_all(char *name, size_t *len)
{
   FILE *f;
   char *buf;
   f= fopen(name,"r");
   if (!f) return NULL;
   read_filep(f, &buf, (size_t*) len);
   fclose(f);
   return buf;
}



aitem_t *amenu_item_new(amenu_t *A, char* name, char *icon,int spos)
{
   aitem_t *T;
   aitem_t *anchor;

   if (spos)
   {
     for(anchor=A->items.next;anchor!=&A->items;anchor=anchor->next)
        if (strcmp(anchor->name, name)>0) break;
   } else anchor= &A->items;

   T= calloc(1,sizeof(*T));
   if (name) T->name= strdup(name);
   if (icon) T->icon= strdup(icon);
   T->prev= anchor->prev;
   T->next= anchor;
   anchor->prev->next= T;
   anchor->prev= T;
   return T;
}

static void destroy_app_menu_item(aitem_t *it)
{
  if (it->name) free(it->name);
  if (it->exec) free(it->exec);
  if (it->src) free(it->src);
  if (it->icon) free(it->icon);
  if (it->sub) destroy_app_menu(it->sub);
  it->prev->next= it->next;
  it->next->prev= it->prev;
  free(it);
}

void destroy_app_menu(amenu_t *A)
{
  while(A->items.next!=&A->items)
  {
    destroy_app_menu_item(A->items.next);
  }
  free(A);
}

amenu_t *amenu_new()
{
  amenu_t *A;
  A= calloc(1,sizeof(*A));
  A->items.prev= A->items.next= &A->items;
  return A;
}

static unsigned char *read_str(unsigned char *K, unsigned char *end, char **R)
{
  unsigned int L;
  *R= "";
  if (end-K<3) return NULL;
  L= K[0]  | (K[1]<<8);
  K+= 2;
  if (K+L>end) return NULL;
  if (!memchr(K, 0, L)) return NULL;
  *R= (char*) K;
  K+= L;
  return K;
}

static unsigned char* read_sub_menu
       (unsigned char *C, unsigned char *end, amenu_t *A, amenu_t *H)
{
  char *name, *icon,*exec, *source;
  aitem_t *T;

  amenu_t *tgt;

#define CHECKC   if (!C) return NULL;
  while(C<end)
  switch(C[0])
  {
  case 'M': 
    C++;
    C= read_str(C,end, &name);         CHECKC
    C= read_str(C,end, &icon);         CHECKC
    T= amenu_item_new( A, name, icon ,0 );
    T->type= 1;
    T->sub= amenu_new();
    C= read_sub_menu(C, end, T->sub, H); CHECKC
    break;
  case 'E':
    C++;
    return C;
  case 'S':
    C++;
    T= amenu_item_new(A, "", "", 0); 
    T->type= 2;
    break;
  case 'I':
  case 'H':
    if (C[0]=='I') tgt= A; else tgt= H;
    C++;
    C= read_str(C, end, &name);    CHECKC
    C= read_str(C, end, &icon);    CHECKC
    C= read_str(C, end, &exec);    CHECKC
    C= read_str(C, end, &source);  CHECKC

    T= amenu_item_new(tgt, name, icon, 0); 
    T->type= 0; T->exec= strdup(exec); T->src= strdup(source);
    break;
  }
  return NULL;
}

amenu_t *read_app_menu(char *fname, amenu_t **Rhidden)
{
  amenu_t *A,*H;
  unsigned char *data; size_t len;
  A= amenu_new(); H= amenu_new();
  data= (unsigned char*) file_read_all(fname, &len);
  if (data) 
  {
    read_sub_menu(data,data+len, A,H);
    free(data);
  }
  if (Rhidden) *Rhidden= H;
          else destroy_app_menu(H);
  return A;
} 

static void write_chr(FILE *f, int x)
{
  char c= x;
  fwrite(&c, 1, 1, f);
}

static void write_str(FILE *f, const char *c)
{
  size_t L;
  if (!c) c= "";
  L= strlen(c);
  if (L>0xfffe) L= 0xfffe;
  write_chr(f, (L+1) & 0xff);
  write_chr(f, (L+1) >> 8 );
  if (L) fwrite(c, 1, L, f);
  write_chr(f, 0);
}

static void write_item(FILE *f, aitem_t *ai, int hid)
{
  aitem_t *si;
  switch(ai->type)
  {
   case 0:
    if (hid) write_chr(f,'H'); 
        else write_chr(f,'I'); 
    write_str(f,ai->name);
    write_str(f,ai->icon);
    write_str(f,ai->exec);
    write_str(f,ai->src); 
    break;
   case 1:
    write_chr(f,'M');
    write_str(f, ai->name);
    write_str(f, ai->icon);
    for(si=ai->sub->items.next;si!=&ai->sub->items;si=si->next)
       write_item(f, si, 0);
    write_chr(f,'E');
    break;
   case 2:
    write_chr(f,'S');
    break;
  }
}


void write_app_menu(const char *fname, amenu_t *menu, amenu_t *hidden)
{
  FILE *f;
  aitem_t *ai;
  f= fopen(fname,"w");
  if (!f) return ;
  for(ai=menu->items.next;ai!=&menu->items;ai=ai->next)
    write_item(f, ai, 0);
  if (hidden)
    for(ai=hidden->items.next;ai!=&hidden->items;ai=ai->next)
      write_item(f, ai, 1);
  write_chr(f,'E');
  fclose(f);
}

