/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-events.h"
#include "x-image.h"
#include "fsel.h"
#include "graymap.h"
#include "chfile.h"
#include "btn.h"

#include "chfile-scr.c"

typedef struct
{
  Window window; int w,h;
  ximage_t *image;
  char *selected;
  int butpos[10];
  int selbut[5];
  void (*func)(void *, char *);
  void *arg;
  file_selector_t *fs;
} fch_t;


static void fc_press(fch_t *fc, XButtonEvent *ev)
{
  btn_click(fc->window, fc->image, fc->butpos, 10, ev, fc->selbut);
}

static void fc_release(fch_t *fc, XButtonEvent *ev)
{
  switch(btn_release(fc->window, fc->image, ev, fc->selbut))
  {
  case 1: if (!fc->selected) return ;
          if (fc->func) (*fc->func)(fc->arg, fc->selected);
  case 2: if (fc->selected) free(fc->selected);
          destroy_image(fc->image);
          unregister_event_handler(fc->window, Expose, NULL);
          unregister_event_handler(fc->window, ButtonPress, NULL);
          unregister_event_handler(fc->window, ButtonRelease, NULL);
          file_selector_free(fc->fs);
          XDestroyWindow(x_display, fc->window);
          free(fc);
          break;
  }
}

static void fc_expose(fch_t *fc, XExposeEvent *ev)
{
  put_image(fc->window, ev->x, ev->y, fc->image,ev->x, ev->y, 
            ev->width, ev->height);
}

static void click_file(void *arg, file_selector_t *fs, char *sel)
{
  fch_t *fc;
  fc= arg;
  if (fc->selected) free(fc->selected);
  fc->selected= strdup(sel);
}

void choose_file(Window parent, int x,int y, int w, int h,
                 char **sfx, char *dir,
                 void (*finish)(void *, char *), void *arg)
{
  XSetWindowAttributes attribs;
  fch_t *fc;
  int i,dx,dy;

  fc= calloc(1,sizeof(*fc));
  fc->image= create_image(w,h);
  fc->w= w;
  fc->h= h;

  fc->func= finish;
  fc->arg= arg;

  attribs.background_pixmap= None;
  fc->window= XCreateWindow(x_display, parent, x, y, w, h,
                            0, x_depth, InputOutput, x_visual,
                            CWBackPixmap, &attribs);
  
  XSelectInput(x_display, fc->window, ExposureMask | KeyPressMask |
                       ButtonPressMask | ButtonReleaseMask);
  register_event_handler(fc->window, Expose, fc, fc_expose, NULL);
  register_event_handler(fc->window, ButtonPress, fc, fc_press, NULL);
  register_event_handler(fc->window, ButtonRelease, fc, fc_release, NULL);

  fill_rect(fc->image->data, fc->w, fc->h, 0,0, fc->w, fc->h, 0xffffff);

  dx= (w - bi_fch_width)/2;
  dy= (h - bi_fch_height);

  blend_alpha_image(fc->image->data, fc->w, fc->h,   dx, dy,
       (uint32_t*)  bi_fch, bi_fch_width, bi_fch_height,
                    0,0, bi_fch_width, bi_fch_height, 0xffffff);
  
  fc->fs= file_selector_new(fc->window, 0,0, w, dy,
                            dir, 0, sfx,
                           click_file, fc);

  i= 0;
  fc->butpos[i++]= dx + 11;
  fc->butpos[i++]= dy + 7;
  fc->butpos[i++]= 64;
  fc->butpos[i++]= 26;
  fc->butpos[i++]= 1; // OK

  fc->butpos[i++]= dx + 78;
  fc->butpos[i++]= dy + 7;
  fc->butpos[i++]= 64;
  fc->butpos[i++]= 26;
  fc->butpos[i++]= 2;  // Cancel

  XMapWindow(x_display, fc->window);
}

