/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-props.h"
#include "x-misc-props.h"
#include "x-events.h"
#include "clients.h"
#include "qumap.h"
#include "resize_image.h"

static int OPT_retain_desk;

client_window_t **client_map;
int n_client_map;



static uint32_t* copy_ulong_to_uint32(unsigned long *data,uint32_t elems)
{
  uint32_t *d,i;
  d= malloc(elems*sizeof(uint32_t));
  for(i=0;i<elems;i++) d[i]= data[i];
  return d;
}

int get_client_icon(client_window_t *cw, uint32_t twidth, uint32_t theight)
{
  unsigned long *idata; 
  size_t n_idata;

  if (cw->flags & CWF_ICON) return 0;
  if (cw->tried & CWF_ICON) return 1;
  cw->tried |= CWF_ICON;
  if (get_window_property_cardinals(cw->id, a_net_wm_icon, &n_idata, &idata,0))
     return 1;
  
  unsigned long *maxi;
  unsigned long *besti,*icon;
  maxi= NULL; besti= NULL;
  for(icon=idata;icon+1<idata+n_idata;icon+= icon[0]*icon[1]+2)
  {
    if (!maxi) 
    { 
      maxi= icon;
    }
    else
    {
      if (icon[0]*icon[1] > maxi[0]*maxi[1]) maxi= icon;
    }
    if (icon[0]>twidth || icon[1]>theight) continue;
    if (!besti) { besti= icon; continue; }
    if (icon[0]*icon[1] > besti[0]*besti[1]) besti= icon;
  }

  if (!besti && !maxi) goto fail;

  if (besti)
  {
    cw->icon_width= besti[0];
    cw->icon_height= besti[1];
    cw->icon_data= copy_ulong_to_uint32(besti+2, besti[0]*besti[1]);
  }
  else
  {
    uint32_t *d;
    d= copy_ulong_to_uint32(maxi+2, maxi[0]*maxi[1]);
    if (resize_image_argb(d, maxi[0], maxi[1],
                          twidth, theight,
                          &cw->icon_data, &cw->icon_width, &cw->icon_height))
    {
       free(d);
       goto fail;
    }
    free(d);
  }
  
  free(idata);
  cw->flags|= CWF_ICON;
  return 0;
fail:
  free(idata); 
  return 1;
}

int get_client_pos(client_window_t *cw)
{
  Window dummy_window;
  int dummy_int; unsigned int dummy_unt;
  unsigned int w,h;

  if (cw->flags & CWF_POS) return 0;
  if (cw->tried & CWF_POS) return 1;
  cw->tried |= CWF_POS;

  XGetGeometry(x_display, cw->id, &dummy_window, &dummy_int, &dummy_int,
                          &w, &h, &dummy_unt, &dummy_unt);

  cw->width= w; cw->height= h;

  XTranslateCoordinates(x_display, cw->id, x_root_window, 0,0,
                        &cw->x, &cw->y, &dummy_window);

  cw->flags|= CWF_POS;
  return 0;
}

int get_client_state(client_window_t *cw)
{
  size_t n_values; Atom *values;
  int i;

  if (cw->flags&CWF_STATE) return 0;
  if (cw->tried&CWF_STATE) return 1;
  cw->tried |= CWF_STATE;
  cw->flags |= CWF_STATE;
  cw->state= 0; // if no state, assume normal

  if (get_window_property_atoms(cw->id, a_net_wm_state, &n_values, &values,0))
    return 0;
  for(i=0;i<n_values;i++)
    if (values[i]==a_net_wm_state_hidden)
    {
      cw->state= 1;
      break;
    } 
  free(values);
  return 0;
}

int get_client_type(client_window_t *cw)
{
  size_t n_values; Atom *values;
  int i;

  if (cw->flags&CWF_TYPE) return 0;
  if (cw->tried&CWF_TYPE) return 1;
  cw->tried |= CWF_TYPE;
  cw->flags |= CWF_TYPE;
  cw->type= 0; // if no type, assume normal

  if (get_window_property_atoms(cw->id, a_net_wm_window_type, 
                                 &n_values, &values,0))
    return 0;

  cw->type= 1;
  
  for(i=0;i<n_values;i++)
    if (values[i]==a_net_wm_window_type_normal)
    {
      cw->type= 0;
      break;
    } 
  for(i=0;i<n_values;i++)
    if (values[i]==a_net_wm_window_type_dock)
    {
      cw->type= 1;
      break;
    }
  free(values);
  return 0;
}

int get_client_desktop_number(client_window_t *cw)
{
  long unsigned int desk;
  unsigned int n;

  if (cw->flags&CWF_DESK) return 0;
  if (cw->tried&CWF_DESK) return 1;

  cw->tried |= CWF_DESK;

  if (get_window_property_cardinal(cw->id, a_net_wm_desktop,
                                     &desk))
    return 1;

  if (OPT_retain_desk && get_number_of_desktops(&n)) n= 1;

  // when the desktop number of a client changes to a
  // sentinel value, then we look at the old value.
  // if the old value was sane, we continue using it.
  // otherwise we simply discard it. this is because
  // blackbox is written in a stupid way and removes
  // minimized windows from their associated desktops.
  if (!OPT_retain_desk || desk<n)
     cw->desk= desk;

  cw->flags |= CWF_DESK;
  return 0;
}

int get_client_title(client_window_t *cw)
{
  size_t slen; char *str;
  if (cw->flags&CWF_TITLE) return 0;
  if (cw->tried&CWF_TITLE) return 1;

  cw->tried |= CWF_TITLE;

  if (get_window_property_utf8_string(cw->id,
               a_net_wm_visible_icon_name, &slen, &str) &&
      get_window_property_utf8_string(cw->id,
               a_net_wm_icon_name, &slen, &str) &&
      get_window_property_utf8_string(cw->id,
               a_net_wm_visible_name, &slen, &str) &&
      get_window_property_utf8_string(cw->id, a_net_wm_name, &slen, &str) &&
      get_window_property_string(cw->id, a_wm_name, &slen, &str) &&
      get_window_property_string(cw->id, a_wm_icon_name, &slen, &str))
   return 1;
  cw->flags |= CWF_TITLE;
  if (cw->title) free(cw->title);
  cw->title= str;
  return 0;
}

/*************************************************************
  The client map is pretty much independent of the other
  functions here.
*************************************************************/

static qumap_t  clients;

static void client_destroyed(void *ignored, XDestroyWindowEvent *ev);

client_window_t *get_client(Window window)
{
  client_window_t *cw;
  unsigned int events;
  XWindowAttributes attribs;

  cw= qumap_find(&clients, window);
  if (cw) return cw;

  cw= calloc(1,sizeof(*cw));
  cw->id= window;
  cw->flags= 0; cw->tried= 0;

  events= StructureNotifyMask;

  if (get_client_type(cw) || cw->type==0) 
  {
    events |= PropertyChangeMask;
    register_event_handler(cw->id, ConfigureNotify, cw,
				   client_size_change,NULL);
    register_event_handler(cw->id, PropertyNotify, cw,
				   client_property_change,NULL);
  }
  XGetWindowAttributes(x_display, window, &attribs);
  events |= attribs.your_event_mask;
  XSelectInput(x_display, cw->id, events);

  register_event_handler(cw->id, DestroyNotify, NULL,
                         client_destroyed, &cw->destroy_handler);

  qumap_insert(&clients, window, cw);
  return cw;
}

static void client_destroyed(void *ignored, XDestroyWindowEvent *ev)
{
  client_window_t *cw;
  cw= qumap_remove(&clients, ev->window);
  if (!cw) return ;
  
  unregister_event_handler(cw->id, DestroyNotify, NULL);
  call_event_handler(cw->destroy_handler, (XEvent*) ev);
  free_event_handler(cw->destroy_handler);

  if (cw->flags&CWF_ICON) free(cw->icon_data);
  if (cw->type==0) 
  {
    unregister_event_handler(cw->id, ConfigureNotify, NULL);
    unregister_event_handler(cw->id, PropertyNotify, NULL);
  }
  if (cw->title) free(cw->title);
  free(cw);
}

void             client_settings(int retain_desk)
{
  OPT_retain_desk= retain_desk;
}
