/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*************************************************************
  Each client window is represented by the following struct.
*************************************************************/

typedef struct 
{
  Window id;
  uint32_t flags, tried;
  int x,y;
  unsigned long width, height;
  uint32_t icon_width, icon_height;
  uint32_t *icon_data;
  int state;  // 0- normal 1- minimized etc.
  int type;   // 0- normal 1- other
  unsigned long desk;
  char *title;
  void *destroy_handler;
} client_window_t;

/***************************************************************
  Any field in this struct could be missing. We store the
  list of present fields in flags. The tried field contains
  the list of fields which we have tried to get, but couldn't.
***************************************************************/

#define CWF_POS    1      // x,y, width, height
#define CWF_ICON   2
#define CWF_STATE  4
#define CWF_TYPE   8
#define CWF_DESK   16
#define CWF_TITLE  32

int get_client_icon(client_window_t *cw, uint32_t twidth, uint32_t theight);
int get_client_pos(client_window_t *cw);
int get_client_state(client_window_t *cw);
int get_client_type(client_window_t *cw);
int get_client_desktop_number(client_window_t *cw);
int get_client_title(client_window_t *cw);

/*************************************************************
  Finally, the current list of clients is held in the 
  global variable 'client_map'.
*************************************************************/

extern client_window_t **client_map; extern int n_client_map;
int get_client_list();


/*************************************************************
  The following function is defined inside the pager module.
  It causes the pager to redraw.
*************************************************************/
void client_size_change(client_window_t *cw, XConfigureEvent *ev);
void client_property_change(client_window_t *cw, XPropertyEvent *ev);

/*************************************************************
  The windows are stored in a map. This function finds the
  corresponding client window struct, or creates one if it 
  doesn't exist.
*************************************************************/
client_window_t *get_client(Window window);
void             forget_client(Window window);


/*************************************************************
  This function initializes the settings regarding the behaviour
  associated with client properties.
*************************************************************/
void             client_settings(int retain_desk);
