/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>

#include "x-basics.h"
#include "x-props.h"
#include "x-image.h"
#include "x-events.h"
#include "graymap.h"
#include "config.h"
#include "menu.h"
#include "imgload.h"
#include "txtren.h"
#include "appmenu.h"
#include "panel.h"

#include "blt-menubutton.xc"

static menu_t *app_menu;
static int startmenu_x;
static Window startmenu_window;
static ximage_t *startmenu_image;
static int parent_y;

/*********************************************************************
  We map integer actions in the menu to actual strings using the 
  following variables:
 ********************************************************************/
static int n_app_entries,s_app_entries;
static char **app_entries;

/*********************************************************************
  The following function registers the given command string into the
  above array.
 ********************************************************************/
static int app_entry(char *k);

static void menu_make(menu_t *M, amenu_t *A);


/*********************************************************************
  We initialize the application menu from the saved configuration. If
  there is no saved configuration, we simply initialize it with the
  configuration program as the only entry.
 ********************************************************************/
static void add_config_program(amenu_t *A);
static void exec_app_entry(int n);

static void startmenu_expose(void *ignore, XExposeEvent *ev);
static void startmenu_click(void *ignore,XButtonEvent *ev_butt);
static void appmenu_done(void *ignore, XAnyEvent *ev);

void init_startmenu
   (Window parent,int panel_y,int panel_h,
    char *menu_fname, int *Rw, Window *Rwin,uint32_t fg,uint32_t bg)
{
  // unsigned int mb_w, mb_h, *mb_data;

  int IW, IH;

  IW= bi_menubutton_width;
  IH= bi_menubutton_height;

  parent_y= panel_y;

  // load_builtin_icon(":menubutton.png", &mb_w, &mb_h, &mb_data);

  startmenu_image= create_image(IW, panel_h);
  fill_rect(startmenu_image->data, IW, panel_h, 0,0, IW, panel_h, bg);

  graymap_draw( startmenu_image->data, IW, panel_h,
                0, (panel_h-IH)/2,
                bi_menubutton,  IW, IH,
                0, 0, fg, bg, IW, IH);

  reload_start_menu(menu_fname);

  Window win;

  XSetWindowAttributes attribs;
  attribs.background_pixmap= None;
  attribs.override_redirect= True;

  win= XCreateWindow(x_display, parent,
                     0, 0, IW, panel_h,
                     0, x_depth, InputOutput, x_visual, 
                    CWBackPixmap | CWOverrideRedirect, &attribs);

  XSelectInput(x_display, win, ExposureMask | ButtonPressMask);
  XMapWindow(x_display, win);

  register_event_handler(win, Expose, NULL, startmenu_expose, NULL);
  register_event_handler(win, ButtonPress, NULL, startmenu_click, NULL);

  startmenu_window= *Rwin= win;
  *Rw= IW;
}



static void menu_make(menu_t *M, amenu_t *A)
{
  unsigned int W,H, *D;
  menu_t *S;
  aitem_t *ai;

  for(ai=A->items.next;ai!=&A->items;ai=ai->next)
  {
    W= H= 0; D= NULL;
    if (ai->type!=2) load_icon(ai->icon, 16, 16, &W, &H, &D);
    switch(ai->type)
    {
    case 0:
      menu_add_item(M, W,H, D, ai->name, app_entry(ai->exec));
      break;
    case 1:
      S= menu_start();
      menu_add_sub(M, W,H, D, ai->name, S);
      menu_make(S, ai->sub);
      menu_end(S);
      break;
    case 2:
      menu_add_sep(M);
      break;
    }
    if (D && ai->icon[0]!=':') free(D);
  }
}

static void add_config_program(amenu_t *A)
{
  aitem_t *T;
  T= amenu_item_new(A, "MuPanel Settings", ":self-settings.png",0);
  T->type= 0;
  T->exec= strdup(":configure");
  T->src= strdup(":internal");
}


static void startmenu_click(void *ignore,XButtonEvent *ev_butt)
{
  Window mw; unsigned int me;
  if (ev_butt->button!=1) return ;
  menu_display_top(app_menu, x_root_height-parent_y, startmenu_x, &mw, &me);
  register_event_handler(mw, me, NULL, appmenu_done, NULL);
}

static void appmenu_done(void *ignore, XAnyEvent *ev)
{
  unregister_event_handler(ev->window, ev->type, NULL);
  exec_app_entry(menu_get_selected(app_menu));
}

static void startmenu_expose(void *ignore, XExposeEvent *ev)
{
  put_image(startmenu_window, ev->x, ev->y,
            startmenu_image, ev->x, ev->y, ev->width, ev->height);
}


static int app_entry(char *k)
{
  int R;
  if (n_app_entries==s_app_entries)
  {
    if (s_app_entries==0) 
    {
      s_app_entries= 100;
      app_entries= malloc(sizeof(app_entries[0])*s_app_entries);
    }
    else
    {
      s_app_entries+= s_app_entries/2;
      app_entries= realloc(app_entries,sizeof(app_entries[0])*s_app_entries);
    }
  }
  R= n_app_entries;
  n_app_entries++;
  app_entries[R]= strdup(k);
  return R;
}

static void exec_app_entry(int n)
{
  if (n<0 || n>=n_app_entries) return ;
  execute_program(app_entries[n]);
}


void reload_start_menu(char *fname)
{
  amenu_t *A;
  if (app_menu) menu_free(app_menu);
  A= read_app_menu(fname,NULL);
  if (A->items.next==&A->items) add_config_program(A);
  app_menu= menu_start();
  menu_make(app_menu, A);
  menu_end(app_menu);
  destroy_app_menu(A);
}
