/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <unistd.h>
#include "appmenu.h"
#include "filerall.h"

static char **icon_path;
static int n_icon_path;

static int has_suffix(char *name, char *sfx)
{
  int Ln, Ls;
  Ln= strlen(name);
  Ls= strlen(sfx);
  if (Ln<Ls) return 0;
  return !strcmp(name+Ln-Ls, sfx);
}

static int ispc(unsigned int K)
{
  if (K==0) return 0;
  if (K<=' ') return 1;
  if (K==127) return 1;
  return 0;
}

static void remove_leading_trailing_spaces(char *K)
{
  unsigned char *L;
  size_t sL;
  L= (unsigned char*) K;
  
  sL= strlen(K);
  while(sL && ispc(L[sL-1])) L[--sL]= 0;

  while(ispc(L[0])) memmove(L, L+1, sL--);
}

static char *get_line(char **src)
{
  char *S;
  char *L;
  S= *src;
  if (!S || !S[0]) return NULL;
  L= S;
  while(*S!=0 && *S!='\n' && *S!='\r') S++;
  switch(*S)
  {
  case 0: *src= NULL; break;
  case '\r': S[0]= 0;
             if (S[1]=='\n') *src= S+2; 
             else *src= S+1; break;
  case '\n': S[0]= 0; *src= S+1;
  }
  return L;
}

static int split_key_value(char *src, char **key, char **value)
{
  char *eq;
  eq= strchr(src,'=');
  if (!eq) return 1;
  *key= src;
  eq[0]= 0;
  *value= eq+1;
  remove_leading_trailing_spaces(*key);
  remove_leading_trailing_spaces(*value);
  return 0;
}

static char **split_semicolon(char *data)
{
  int N,i;
  char *K,**R;
  for(K=data,N=0;K;)
  {
    N++;
    K= strchr(K,';');
    if (!K) break;
    K++; 
  }
  R= malloc(sizeof(char*)*(N+1));
  i= 0;
  while(data[0])
  {
    K= strchr(data,';');
    if (!K) K=data+strlen(data);
    R[i]= malloc(K-data+1);
    memcpy(R[i], data, K-data);
    R[i][K-data]= 0;
    remove_leading_trailing_spaces(R[i]);
    if (R[i][0]) i++;
            else free(R[i]);
    data= K; if (data[0]==';') data++;
  }
  R[i]= NULL;
  return R;
}

static const char *get_app_category(char **categories)
{
/* This function finds the top level menu the application should go
   into. We do this only if
    - We don't find the application in the menu already, and
    - We don't find the application's category in the menu

   Unfortunately, applications have more than one category. So, I have
   to map them to menu entries manually.

   In order to get the list of all categories in all applications, run
   the following:

   grep Categor /usr/share/applications/ *.desktop \
    | gawk -F '=' '{print $2}' | \
      gawk -F ';' '{ for(i=1;i<=NF;i++) { printf("\"%s\",\n",$i); } }' \
      | sort  | uniq 
*/
   static const char *catmap[]= {
".Other",
".Graphics",    "Graphics", "2DGraphics", "VectorGraphics",
                "RasterGraphics",
".Office",      "Office", "Spreadsheet", "WordProcessor", "Presentation",
                "Scanning",
                "X-MandrivaLinux-Office-Drawing",
                "X-MandrivaLinux-Office-Other",
                "X-MandrivaLinux-Office-Presentations",
                "X-MandrivaLinux-Office-Spreadsheets",
                "X-MandrivaLinux-Office-Wordprocessors",
".Development", "Development", "GUIDesigner", "Java",
".Education",   "Education",
".Electronics", "Electronics",
".Accessories", "Calculator", "Clock", "Compression", "FileManager",
                "FileTools", "FileTransfer", "Filesystem", "Dictionary",
                "TerminalEmulator", "TextEditor", "Utility", "Archiving",
".Games",       "Game", "KidsGame", "LogicGame",
".Multimedia",  "Video", "Player", "Audio", "AudioVideo", "Recorder",
                "X-AudioVideoImport",
".Internet",    "WebBrowser", "P2P",
".Settings",    "Settings", "HardwareSettings", "DesktopSettings",
                "X-XFCE-HardwareSettings", "X-XFCE-PersonalSettings",
                "X-XFCE-SettingsDialog", "X-XFCE-SystemSettings",
                "X-Xfce-Toplevel", "X-GNOME-NetworkSettings",
                "X-GNOME-PersonalSettings", "X-GNOME-Settings-Panel",
                "X-GNOME-SystemSettings", "X-LXDE-Settings",
   NULL };
   int i,j;
   for(i=0;categories[i];i++)
     for(j=0;catmap[j];j++)
       if (!strcmp(catmap[j], categories[i]))
       {
         while(catmap[j][0]!='.') j--;
         return catmap[j]+1;
       }
   return catmap[0]+1;
}

static int find_app_in_menu(amenu_t *A, char *source,char *exec)
{
  aitem_t *ai;
  for(ai=A->items.next;ai!=&A->items;ai=ai->next)
    switch(ai->type)
    {
    case 1:
        if (find_app_in_menu(ai->sub, source, exec)) 
           return 1;
        break;
    case 0:
        if (strcmp(ai->src,source)) continue;
        if (ai->exec) free(ai->exec);
        ai->exec= strdup(exec);
        return 1;
    } 
  return 0;
}

static amenu_t *find_appcat_in_menu(amenu_t *A, char **categories)
{
  aitem_t *ai;
  amenu_t *R;
  int i;
  for(ai=A->items.next;ai!=&A->items;ai=ai->next)
    if (ai->type==1)
    {
      for(i=0;categories[i];i++)
        if (!strcmp(ai->name, categories[i])) return ai->sub;
      R= find_appcat_in_menu(ai->sub, categories); 
      if (R) return R;
    }
  return NULL;
}

static char *get_cat_icon(char *name)
{
  static const char *list[]=
  {
"Other",       ":cat-other.png",
"Graphics",    ":cat-graphics.png",
"Office",      ":cat-office.png",
"Development", ":cat-development.png",
"Accessories", ":cat-accessories.png",
"Games",       ":cat-games.png",
"Multimedia",  ":cat-multimedia.png",
"Internet",    ":cat-internet.png",
"Settings",    ":cat-settings.png",
NULL
  };
  int i;
  for(i=0;list[i];i+=2)
     if (!strcmp(list[i],name)) return (char*) list[i+1];
  return "";
}

static amenu_t *create_appcat_in_menu(amenu_t *A, char **categories)
{
  amenu_t *R;
  char *name;
  aitem_t *ai;

  R= find_appcat_in_menu(A, categories);
  if (R) return R;
  name= (char*) get_app_category(categories);
  for(ai=A->items.next;ai!=&A->items;ai=ai->next)
    if (ai->type==1 && !strcmp(ai->name,name))
       return ai->sub;

  ai= amenu_item_new(A,name, get_cat_icon(name), 1);
  ai->type= 1;
  ai->sub= amenu_new();
  return ai->sub;
}

static void icon_path_with_size
  (char *buf, const char *path, int size, char *name)
{
  sprintf(buf,"%s/%dx%d/apps/%s.png", path, size, size, name);
}

static char *find_icon(char *icon,int icon_size)
{
  char buf[2000];
  int i,k;
  int best_path, best_size;
  best_size= 0; best_path= 0;
  static unsigned int sizes[]= { 16,22, 24, 32, 36, 48, 64, 72, 96, 128,
                                 192, 256, 0 };
  if (icon[0]==':' || strchr(icon,'/')) return strdup(icon);
  for(i=0;i<n_icon_path;i++)
    for(k=0;sizes[k];k++)
    {
      if (best_size==icon_size) goto done;
      icon_path_with_size(buf, icon_path[i], sizes[k], icon);
      if (access(buf,R_OK)) continue;
      if (sizes[k]>best_size)
            { best_size= sizes[k]; best_path= i; }
    }
done:
  if (best_size==0)
     return strdup(icon);
  icon_path_with_size(buf, icon_path[best_path], best_size, icon);
  return strdup(buf);
}

static void free_strings(char **strs)
{
  int i;
  if (!strs) return ;
  for(i=0;strs[i];i++)
    free(strs[i]);
  free(strs);
}

static void scan_app_dir(char *dirname, amenu_t *menu, amenu_t *hidden)
{
   DIR *apps;
   char *K,*L;
   char *key, *value;
   struct dirent *diren;
   apps= opendir(dirname);

   if (!apps) return ;

   while((diren=readdir(apps)))
   {
     char bufname[2000];
     char *data; size_t len;
     if (!has_suffix(diren->d_name,".desktop")) continue;
     sprintf(bufname,"%s/%s",dirname, diren->d_name);
     data= file_read_all(bufname, &len);
     if (!data) continue;
     K= L= data;
     // printf("file %s\n", diren->d_name);
     char *name, *category, *showin, *icon, *exe, *nodis;
     nodis= name= category= showin= icon= exe= NULL;
     while((L=get_line(&K)))
     {
        remove_leading_trailing_spaces(L);
        if (split_key_value(L, &key, &value)) continue;
        if (!strcmp(key,"Name"))            name= value;
        else if (!strcmp(key,"Categories")) category= value;
        else if (!strcmp(key,"OnlyShowIn")) showin= value;
        else if (!strcmp(key,"Icon"))       icon= value; 
        else if (!strcmp(key,"Exec"))       exe= value;
        else if (!strcmp(key,"NoDisplay"))  nodis= value;
     }
     if (nodis && !strcasecmp(nodis,"True"))  continue;
     
     if (exe && 
           !find_app_in_menu( menu,bufname, exe) &&
           !find_app_in_menu(hidden, bufname, exe)) 
     {
        char **cats;
        amenu_t *tgt;
        aitem_t *T;
        if (!name) name= exe;
        if (!category) category= "Other";
        cats= split_semicolon(category);
        tgt= create_appcat_in_menu(menu, cats);
        if (icon) icon= find_icon(icon,16);
        T= amenu_item_new(tgt, name, icon , 1);
        T->type= 0;
        T->exec= strdup(exe);
        T->src= strdup(bufname);
        if (icon) free(icon);
        free_strings(cats);
     }
     free(data);
   }
   closedir(apps);
}

void scan_app_dirs(char **appdirs, int n_appdirs,
                   char **icondirs, int n_icondirs,
                   amenu_t *menu, amenu_t *hidden)
{
   int i;
   icon_path= icondirs;
   n_icon_path= n_icondirs;

   for(i=0;i<n_appdirs;i++)
     scan_app_dir(appdirs[i], menu, hidden);
}
