/*
Copyright (c) 2014-2016 Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#line 410 "txtren.u"
#include <ft2build.h>
#include FT_FREETYPE_H
#include  <freetype/ftglyph.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "txtren.h"

#line 36 "txtren.u"

typedef struct font_file
{
  int mapped;
  int refcount;
  char *name;
  const unsigned char *data;
  size_t len;
  struct font_file *next;
} font_file_t;

#line 150 "txtren.u"

typedef struct k_font
{
  u_font_t user;
  int ptsize;
  FT_Face face;
  int refcount;
  font_file_t *file;
  struct k_font *next;
} k_font_t;

#line 14 "txtren.u"

static int font_hor_dpi=96, font_ver_dpi=96;

#line 52 "txtren.u"

static font_file_t *font_files;
static font_file_t *fallback_font_file;

#line 162 "txtren.u"

static k_font_t *fonts;

#line 442 "txtren.u"

static FT_Library ftlib;

#line 57 "txtren.c"

static font_file_t *font_load_file(const char *name);
static void unload_font_file(font_file_t *ff);
static void compute_font_metrics(u_font_t *uf, FT_Face face, int ptsize);
 
static int utf8_decode_string
  (const uint8_t *input, size_t ilen, uint32_t *output, size_t *olen);
#line 429 "txtren.u"

#line 75 "txtren.u"

static font_file_t *font_load_file(const char *name)
{
  font_file_t *ff;
  int fd;
  struct stat sbuf;
  char *mapped;

  for(ff=font_files;ff;ff=ff->next)
    if (!strcmp(ff->name, name)) return ff;
  if (name[0]==':') return NULL;
  
  fd= open(name, O_RDONLY);
  if (fd<0) return NULL;
  if (fstat(fd, &sbuf)) goto fail;

  mapped= mmap(NULL, sbuf.st_size, PROT_READ, MAP_SHARED, fd, 0);
  if (mapped==MAP_FAILED) goto fail;

  ff= calloc(1,sizeof(*ff));
  ff->name= strdup(name);
  ff->data= mapped;
  ff->len= sbuf.st_size;
  ff->next= font_files;
  font_files= ff;
  ff->mapped= 1;

  close(fd);
  return ff;

fail:
  close(fd);
  return NULL;
}

#line 114 "txtren.u"

static void unload_font_file(font_file_t *ff)
{
  size_t len;
  long page;
  font_file_t *y,**L;
  if (!ff->mapped) return ;

  for(y=font_files, L= &font_files;y;L=&y->next,y=y->next)
     if (y==ff) break;
  if (!y) return ;
  *L= y->next;

  page= sysconf(_SC_PAGE_SIZE);
  if (page<=0) page=4096;
  len= (ff->len+page-1)/page;
  len *= page;
  munmap((void*) ff->data, len);
  free(ff->name);
  free(ff);
}

#line 268 "txtren.u"

static void compute_font_metrics(u_font_t *uf, FT_Face face, int ptsize)
{
  double v_scale;
  if (ptsize<1) ptsize= 1;
  v_scale= ((double) face->units_per_EM*72) / ((double)ptsize*font_ver_dpi);
  uf->ascent= ceil( face->bbox.yMax / v_scale );
  uf->descent= ceil( -face->bbox.yMin / v_scale );

  if (uf->ascent<1) uf->ascent= 1;
  if (uf->descent<1) uf->descent= 1;
  uf->height= uf->ascent + uf->descent;
}

#line 462 "txtren.u"


/*
UTF-8 Table
8421 8421
110x xxxx   2 bytes    mask= e0  result= c0   min= 80       vmask= 1f
1110 xxxx   3 bytes    mask= f0  result= e0   min= 800      vmask= 0f
1111 0xxx   4 bytes    mask= f8  result= f0   min= 10000    vmask= 07
1111 10xx   5 bytes    mask= fc  result= f8   min= 200000   vmask= 03
1111 110x   6 bytes    mask= fe  result= fc   min= 4000000  vmask= 01

10xx xxxx   cont. byte mask= c0  result= 80   vmask= 3f
*/
static int utf8_decode_string
  (const uint8_t *input, size_t ilen, uint32_t *output, size_t *olen)
{
   const uint8_t *end,*start;
   static const uint32_t bit31= (1<<31);
   size_t L;
   int have_errors;
   int i, contb;
   uint8_t C;
   uint32_t outc; uint32_t min;

   end= input+ilen;
   L= 0;
   have_errors= 0;

   if (ilen==0) goto done;

again:
   C= input[0];
   if ( (C&0x80) ==0)    { outc= C; input++; goto out; }
   if ( (C&0xe0) ==0xc0) goto seq2; 
   if ( (C&0xf0) ==0xe0) goto seq3;
   if ( (C&0xf8) ==0xf0) goto seq4;

   // Here, we don't care about the seq5 and seq6 stuff.
   // they are invalid anyway, either overlong or greater
   // then the unicode limit. We just treat them as unrecognized
   // bytes and output them.
invalid_byte:
   outc= input[0] | bit31;
   input++;
   have_errors= 1;
   goto out;

seq2: outc= C & 0x1f;  min=    0x80; contb= 1; goto do_cont;
seq3: outc= C & 0x0f;  min=   0x800; contb= 2; goto do_cont;
seq4: outc= C & 0x07;  min= 0x10000; contb= 3; goto do_cont;

do_cont:
   start= input;
   input++;
   for(i=0;i<contb;i++)
   {
     if ((input==end) || ( (input[0]&0xc0) !=0x80)) goto reject;
     outc= (outc << 6) | (input[0]&0x3f);
     input++;
   }

   if (outc<min || outc>0x10ffff) goto reject; 
   if (outc>=0xd800 && outc<=0xdfff) goto reject;
   if (outc==0xffff || outc==0xfffe) goto reject;

out:
   output[L++]= outc;
   if (input<end) goto again;
done:
   *olen= L;
   return have_errors;
reject:
   input= start;
   goto invalid_byte; 
}


#line 18 "txtren.u"

void text_set_dpi(int hor, int ver)
{
  if (hor<=0 || ver<=0) return ;
  font_hor_dpi= hor;
  font_ver_dpi= ver;
}

#line 57 "txtren.u"

void font_builtin(const char *name, const unsigned char *data, size_t len)
{
  font_file_t *ff;
  ff= calloc(1,sizeof(*ff));
  ff->name= strdup(name);
  ff->data= data;
  ff->len= len;
  ff->next= font_files;
  font_files= ff;
  if (!fallback_font_file) 
     fallback_font_file= ff;
}
#line 170 "txtren.u"

u_font_t *font_new(const char *name, int ptsize)
{
  k_font_t *kf;
  font_file_t *ff;
  FT_Error ftresult;
  FT_Face face;

  for(kf=fonts;kf;kf=kf->next)
    if (kf->ptsize==ptsize && !strcmp(kf->file->name,name))
    {
       kf->refcount++;
       return &kf->user; 
    }

  ff= font_load_file(name);
  if (!ff) ff= fallback_font_file;
again:
  if (!ff) return NULL; 

  ftresult= FT_New_Memory_Face(ftlib, ff->data, ff->len, 0, &face);
  if (ftresult) goto try_default;

  ftresult= FT_Set_Char_Size(face, 0, ptsize<<6, font_hor_dpi, font_ver_dpi);
  if (ftresult) goto fail_set_size;

  kf= calloc(1,sizeof(*kf));
  kf->user.priv= kf;
  kf->ptsize= ptsize;
  kf->face= face;
  kf->refcount= 1;
  kf->file= ff;
  kf->next= fonts;

  fonts= kf;
  ff->refcount++;

  compute_font_metrics(&kf->user, face, ptsize);
  return kf->user.priv;

fail_set_size:
  FT_Done_Face(face);
try_default:
  if (ff!=fallback_font_file) 
  {
    unload_font_file(ff);
    ff=fallback_font_file;
    goto again;
  }
  return NULL;
}

#line 287 "txtren.u"

void font_free(u_font_t *font)
{
  k_font_t *y,**L,*kf;

  kf= font->priv;
  kf->refcount--;
  if (kf->refcount>0) return ;

  for(y=fonts, L=&fonts;y;L=&y->next, y=y->next)
    if (y==kf) 
       break;
  if (!y) return ;
  *L= y->next;

  kf->file->refcount--;
  if (kf->file->refcount<=0)
     unload_font_file(kf->file);
  FT_Done_Face(kf->face);
  free(kf);
}

#line 310 "txtren.u"

void text_render(u_font_t *font, const char *str, 
                 uint32_t *Rw, uint8_t **Rd)
{
  uint32_t *wide_str;
  size_t   wide_str_len, str_len;
  str_len= strlen(str);
  wide_str= malloc(sizeof(uint32_t) * (str_len+1));
  utf8_decode_string((const uint8_t*) str, str_len, wide_str, &wide_str_len);
  text_render_wide(font, wide_str, wide_str_len, Rw, Rd);
  free(wide_str);
}

void text_render_wide(u_font_t *font, uint32_t *wide_str,
                 size_t wide_str_len,
                 uint32_t *Rw, uint8_t **Rd)
{
  k_font_t *kf;
  uint32_t *pos_x;
  uint32_t glyph_index;
  FT_Glyph *glyphs;
  FT_Face face;
  FT_BitmapGlyph glyph_bitmap;           

  uint32_t r,c,i,x,n_glyphs;
  int rows,cols, left_padding, top_padding;
  int dst_y, dst_x;
  int W;
  uint8_t *data, *buffer;

  kf= font->priv;
  face= kf->face;


  pos_x= malloc(sizeof(pos_x[0])*wide_str_len);
  glyphs= malloc(sizeof(glyphs[0])*wide_str_len);
  n_glyphs= 0;
  x= 0;

  for(i=0;i<wide_str_len;i++)
  {
    if (wide_str[i]&(1<<31))
       glyph_index= 0;
    else 
       glyph_index= FT_Get_Char_Index(face, wide_str[i]);

     if (FT_Load_Glyph(face,glyph_index, FT_LOAD_RENDER) ||
         FT_Get_Glyph(face->glyph, &glyphs[n_glyphs])) continue;
     pos_x[n_glyphs]= x;
     x+= glyphs[n_glyphs]->advance.x >> 16;
     n_glyphs++;
  }

  if (x<1) x= 1;
  W= x;
  data= calloc(1, kf->user.height * W );

  for(i=0;i<n_glyphs;i++)
  {
     if (glyphs[i]->format == FT_GLYPH_FORMAT_BITMAP)
     {
       glyph_bitmap= (FT_BitmapGlyph) glyphs[i];

       rows= glyph_bitmap->bitmap.rows;
       cols= glyph_bitmap->bitmap.width;
       buffer= glyph_bitmap->bitmap.buffer;

       left_padding= glyph_bitmap->left;
       if (left_padding<0) left_padding= 0;

       top_padding= (kf->user.ascent - glyph_bitmap->top);
       if (top_padding<0) top_padding= 0;

       for(r=0;r<rows;r++)
       {
         dst_y= r+top_padding; 
         if (dst_y<kf->user.height) 
           for(c=0;c<cols;c++)
           {
             dst_x= (pos_x[i] + left_padding + c);
             if (dst_x<0 || dst_x>=W) continue;
             data[ dst_y * W + dst_x ] = buffer[c];
           }
         buffer+= glyph_bitmap->bitmap.pitch;
       }
     }
     FT_Done_Glyph(glyphs[i]);                                
  }
  free(pos_x);
  free(glyphs);

  *Rw= W;
  *Rd= data;
}


#line 446 "txtren.u"

int font_init()
{
   return FT_Init_FreeType(&ftlib);
}

int font_finish()
{
   FT_Done_Library(ftlib);
}

#line 432 "txtren.u"

